@@ -1,5 +1,5 @@
 ;; shepherd.scm -- The daemon shepherd.
-;; Copyright (C) 2013-2014, 2016, 2018-2020, 2022 Ludovic Courtès <ludo@gnu.org>
+;; Copyright (C) 2013-2014, 2016, 2018-2020, 2022-2023 Ludovic Courtès <ludo@gnu.org>
 ;; Copyright (C) 2002, 2003 Wolfgang Jährling <wolfgang@pro-linux.de>
 ;; Copyright (C) 2018 Carlo Zancanaro <carlo@zancanaro.id.au>
 ;; Copyright (C) 2018 Danny Milosavljevic <dannym@scratchpost.org>
@@ -30,6 +30,7 @@
   #:use-module (srfi srfi-1)    ;; List library.
   #:use-module (srfi srfi-26)
   #:use-module (srfi srfi-34)
+  #:use-module (srfi srfi-35)
   #:use-module (shepherd config)
   #:use-module (shepherd support)
   #:use-module (shepherd service)
@@ -41,17 +42,6 @@
 
 (define sleep (@ (fibers) sleep))
 
-(define (open-server-socket file-name)
-  "Open a socket at FILE-NAME, and listen for connections there."
-  (with-fluids ((%default-port-encoding "UTF-8"))
-    (let ((sock    (socket PF_UNIX
-                           (logior SOCK_STREAM SOCK_NONBLOCK SOCK_CLOEXEC)
-                           0))
-          (address (make-socket-address AF_UNIX file-name)))
-      (bind sock address)
-      (listen sock 10)
-      sock)))
-
 (define-syntax-rule (unwind-protect body ... conclude)
   "Evaluate BODY... and return its result(s), but always evaluate CONCLUDE
 before leaving, even if an exception is raised.
@@ -85,7 +75,7 @@ socket file at FILE-NAME upon exit of PROC.  Return the values of PROC."
                                    (strerror (system-error-errno args)))
                      ;; Stop services that were started from the config file
                      ;; and quit.
-                     (stop 'root)))))))
+                     (stop-service root-service)))))))
     (unwind-protect (proc sock)
                     (begin
                       (close sock)
@@ -132,7 +122,7 @@ already ~a threads running, disabling 'signalfd' support")
    (lambda ()
      (catch 'quit
        (lambda ()
-         (stop root-service))
+         (stop-service root-service))
        quit-exception-handler))))
 
 (define (signal-handler signal)
@@ -165,7 +155,7 @@ already ~a threads running, disabling 'signalfd' support")
            (fcntl fd F_SETFD (logior FD_CLOEXEC flags)))))
       (loop (+ fd 1)))))
 
-(define* (run-daemon #:key (config-file (default-config-file)) persistency
+(define* (run-daemon #:key (config-file (default-config-file))
                      socket-file pid-file signal-port poll-services?)
   ;; We might have file descriptors inherited from our parent, as well as file
   ;; descriptors wrongfully opened by Guile or Fibers (see
@@ -183,15 +173,6 @@ already ~a threads running, disabling 'signalfd' support")
     (lambda (key . args)
       (caught-error key args)
       (quit 1)))
-  ;; Start what was started last time.
-  (and persistency
-       (catch 'system-error
-         (lambda ()
-           (start-in-order (read (open-input-file
-                                  persistency-state-file))))
-         (lambda (key . args)
-           (apply format #f (gettext (cadr args)) (caddr args))
-           (quit 1))))
 
   ;; Ignore SIGPIPE so that we don't die if a client closes the connection
   ;; prematurely.
@@ -244,6 +225,28 @@ already ~a threads running, disabling 'signalfd' support")
 
            (next-command))))))
 
+(define-syntax replace-core-bindings!
+  (syntax-rules (<>)
+    "Replace the given core bindings in the current process, restoring them upon
+fork in the child process."
+    ((_ () <> ((binding value) ...))
+     (let ((real-primitive-fork primitive-fork))
+       (set! primitive-fork
+             (lambda ()
+               (let ((result (real-primitive-fork)))
+                 (when (zero? result)
+                   (set! binding value)
+                   ...
+                   (set! primitive-fork real-primitive-fork))
+                 result)))))
+    ((_ ((binding value) rest ...) <> (saved-bindings ...))
+     (let ((real binding))
+       (set! binding value)
+       (replace-core-bindings! (rest ...) <>
+                               ((binding real) saved-bindings ...))))
+    ((_ (binding value) ...)
+     (replace-core-bindings! ((binding value) ...) <> ()))))
+
 
 ;; Main program.
 (define (main . args)
@@ -286,65 +289,56 @@ already ~a threads running, disabling 'signalfd' support")
     ;; Process command line arguments.
     (process-args (program-name) args
 		  ""
-		  "This is a service manager for Unix and GNU."
+		  (l10n "This is a service manager for Unix and GNU.")
 		  not ;; Fail on unknown args.
-		  (make <option>
-		    #:long "persistency" #:short #\p
-		    #:takes-arg? #t #:optional-arg? #t
-                    #:arg-name (l10n "FILE")
-		    #:description (l10n "use FILE to load and store state")
-		    #:action (lambda (file)
-			       (set! persistency #t)
-			       (and file
-				    (set! persistency-state-file file))))
-		  (make <option>
-		    #:long "quiet"
-		    #:takes-arg? #f
+		  (option
+		    #:long-name "quiet"
+		    #:takes-argument? #f
 		    #:description (l10n "synonym for --silent")
 		    #:action (lambda ()
                                ;; XXX: Currently has no effect.
                                #t))
-		  (make <option>
-		    #:long "silent" #:short #\S
-		    #:takes-arg? #f
+		  (option
+		    #:long-name "silent" #:short-name #\S
+		    #:takes-argument? #f
 		    #:description (l10n "don't do output to stdout")
 		    #:action (lambda ()
                                ;; XXX: Currently has no effect.
                                #t))
-		  (make <option>
+		  (option
 		    ;; It might actually be desirable to have an
 		    ;; ``insecure'' setup in some circumstances, thus
 		    ;; we provide it as an option.
-		    #:long "insecure" #:short #\I
-		    #:takes-arg? #f
+		    #:long-name "insecure" #:short-name #\I
+		    #:takes-argument? #f
 		    #:description (l10n "don't ensure that the setup is secure")
 		    #:action (lambda ()
                                (set! secure #f)))
-		  (make <option>
-		    #:long "logfile" #:short #\l
-		    #:takes-arg? #t #:optional-arg? #f
-                    #:arg-name (l10n "FILE")
+		  (option
+		    #:long-name "logfile" #:short-name #\l
+		    #:takes-argument? #t #:argument-is-optional? #f
+                    #:argument-name (l10n "FILE")
 		    #:description (l10n  "log actions in FILE")
 		    #:action (lambda (file)
 			       (set! logfile file)))
-		  (make <option>
-		    #:long "pid"
-		    #:takes-arg? #t #:optional-arg? #t
-                    #:arg-name (l10n "FILE")
+		  (option
+		    #:long-name "pid"
+		    #:takes-argument? #t #:argument-is-optional? #t
+                    #:argument-name (l10n "FILE")
 		    #:description (l10n "when ready, write PID to FILE or stdout")
 		    #:action (lambda (file)
 			       (set! pid-file (or file #t))))
-		  (make <option>
-		    #:long "config" #:short #\c
-		    #:takes-arg? #t #:optional-arg? #f
-                    #:arg-name (l10n "FILE")
+		  (option
+		    #:long-name "config" #:short-name #\c
+		    #:takes-argument? #t #:argument-is-optional? #f
+                    #:argument-name (l10n "FILE")
 		    #:description (l10n "read configuration from FILE")
 		    #:action (lambda (file)
 			       (set! config-file file)))
-		  (make <option>
-		    #:long "socket" #:short #\s
-		    #:takes-arg? #t #:optional-arg? #f
-                    #:arg-name (l10n "FILE")
+		  (option
+		    #:long-name "socket" #:short-name #\s
+		    #:takes-argument? #t #:argument-is-optional? #f
+                    #:argument-name (l10n "FILE")
 		    #:description
 		    (l10n "get commands from socket FILE or from stdin (-)")
 		    #:action (lambda (file)
@@ -369,7 +363,7 @@ already ~a threads running, disabling 'signalfd' support")
                     (if (and (not logfile) (zero? (getuid)))
                         (format #f "shepherd[~d]: " (getpid))
                         default-logfile-date-format))
-                   (current-output-port
+                   (%current-service-output-port
                     ;; Send output to log and clients.
                     (make-shepherd-output-port
                      (if (and (zero? (getuid)) (not logfile))
@@ -379,93 +373,70 @@ already ~a threads running, disabling 'signalfd' support")
                          (%make-void-port "w")
                          (current-output-port)))))
 
-      (set-port-encoding! (log-output-port) "UTF-8")
+      (parameterize ((current-output-port (%current-service-output-port)))
+        (set-port-encoding! (log-output-port) "UTF-8")
 
-      ;; Start the 'root' service.
-      (start root-service)
+        (when (= 1 (getpid))
+          ;; When running as PID 1, disable hard reboots upon ctrl-alt-del.
+          ;; Instead, the kernel will send us SIGINT so that we can gracefully
+          ;; shut down.  See ctrlaltdel(8) and kernel/reboot.c.
+          (catch 'system-error
+            (lambda ()
+              (disable-reboot-on-ctrl-alt-del))
+            (lambda args
+              (let ((err (system-error-errno args)))
+                ;; When in a separate PID namespace, we get EINVAL (see
+                ;; 'reboot_pid_ns' in kernel/pid_namespace.c.)  We get EPERM in
+                ;; a user namespace that lacks CAP_SYS_BOOT.
+                (unless (member err (list EINVAL EPERM))
+                  (apply throw args)))))
 
-      (when (= 1 (getpid))
-        ;; When running as PID 1, disable hard reboots upon ctrl-alt-del.
-        ;; Instead, the kernel will send us SIGINT so that we can gracefully
-        ;; shut down.  See ctrlaltdel(8) and kernel/reboot.c.
-        (catch 'system-error
-          (lambda ()
-            (disable-reboot-on-ctrl-alt-del))
-          (lambda args
-            (let ((err (system-error-errno args)))
-              ;; When in a separate PID namespace, we get EINVAL (see
-              ;; 'reboot_pid_ns' in kernel/pid_namespace.c.)  We get EPERM in
-              ;; a user namespace that lacks CAP_SYS_BOOT.
-              (unless (member err (list EINVAL EPERM))
-                (apply throw args)))))
+          ;; Load the SIGSEGV/SIGABRT handler.  This is what allows PID 1 to
+          ;; dump core on "/", should something go wrong.
+          (false-if-exception
+           (dynamic-link (string-append %pkglibdir "/crash-handler"))))
 
-        ;; Load the SIGSEGV/SIGABRT handler.  This is what allows PID 1 to
-        ;; dump core on "/", should something go wrong.
-        (false-if-exception
-         (dynamic-link (string-append %pkglibdir "/crash-handler"))))
+        ;; Install signal handlers for everything but SIGCHLD, which is taken
+        ;; care of in (shepherd services).
+        (for-each (lambda (signal)
+                    (sigaction signal (signal-handler signal)))
+                  (delete SIGCHLD %precious-signals))
 
-      ;; Install signal handlers for everything but SIGCHLD, which is taken
-      ;; care of in (shepherd services).
-      (for-each (lambda (signal)
-                  (sigaction signal (signal-handler signal)))
-                (delete SIGCHLD %precious-signals))
+        ;; Run Fibers in such a way that it does not create any POSIX thread,
+        ;; because POSIX threads and 'fork' cannot be used together.
+        (run-fibers
+         (lambda ()
+           (with-service-registry
 
-      ;; Run Fibers in such a way that it does not create any POSIX thread,
-      ;; because POSIX threads and 'fork' cannot be used together.
-      (run-fibers
-       (lambda ()
-         (catch 'quit
-           (lambda ()
-             (with-process-monitor
+             ;; Register and start the 'root' service.
+             (register-services (list root-service))
+             (start-service root-service)
 
-               ;; Replace the default 'system*' binding with one that
-               ;; cooperates instead of blocking on 'waitpid'.
-               (let ((real-system* system*))
-                 (set! system* spawn-command)
+             (catch 'quit
+               (lambda ()
+                 (with-process-monitor
+                   ;; Replace the default 'system*' binding with one that
+                   ;; cooperates instead of blocking on 'waitpid'.  Replace
+                   ;; 'primitive-load' (in C as of 3.0.9) with one that does
+                   ;; not introduce a continuation barrier.
+                   (replace-core-bindings!
+                    (system* (lambda command
+                               (spawn-command command)))
+                    (system spawn-shell-command)
+                    (primitive-load primitive-load*))
 
-                 ;; Restore 'system*' after fork.
-                 (set! primitive-fork
-                       (let ((real-fork primitive-fork))
-                         (lambda ()
-                           (let ((result (real-fork)))
-                             (when (zero? result)
-                               (set! primitive-fork real-fork)
-                               (set! system* real-system*))
-                             result)))))
-
-               (run-daemon #:socket-file socket-file
-                           #:config-file config-file
-                           #:pid-file pid-file
-                           #:signal-port signal-port
-                           #:poll-services? poll-services?
-                           #:persistency persistency)))
-           (case-lambda
-             ((key value . _)
-              (primitive-exit value))
-             ((key)
-              (primitive-exit 0)))))
-       #:parallelism 1  ;don't create POSIX threads
-       #:hz 0))))       ;disable preemption, which would require POSIX threads
-
-;; Start all of SERVICES, which is a list of canonical names (FIXME?),
-;; but in a order where all dependencies are fulfilled before we
-;; attempt to start a service.  It is assumed that this is possible (FIXME?).
-;; FIXME: This procedure is untested!
-(define (start-in-order services)
-  ;; Same thing with an association list where the cdr of each pair is
-  ;; `#f' if the service is not yet started.
-  (define (start-in-order-assoc assoc-services)
-    (while (any not (map cdr assoc-services))
-      (for-each (lambda (x)
-		  (let ((service (car (lookup-services (car x)))))
-		    (and (not (cdr x))
-			 (depends-resolved? service)
-			 (set-cdr! x (start service)))))
-		assoc-services)))
-
-  (start-in-order-assoc (map (lambda (x)
-			       (cons x #f))
-			     services)))
+                   (run-daemon #:socket-file socket-file
+                               #:config-file config-file
+                               #:pid-file pid-file
+                               #:signal-port signal-port
+                               #:poll-services? poll-services?)))
+               (case-lambda
+                 ((key value . _)
+                  (primitive-exit value))
+                 ((key)
+                  (primitive-exit 0))))))
+         #:parallelism 1                          ;don't create POSIX threads
+         #:hz 0)))))       ;disable preemption, which would require POSIX threads
 
 (define (process-connection sock)
   "Process client connection SOCK, reading and processing commands."
@@ -498,52 +469,86 @@ already ~a threads running, disabling 'signalfd' support")
         (local-output (l10n "Exiting."))
         (primitive-exit 0))))              ;leave without going through Fibers
 
+(define (call-with-command-message-port command proc)
+  "Call @var{proc} passing it a procedure to retrieve the messages emitted
+while evaluating @var{command}."
+  (define message-port
+    (with-fluids ((%default-port-encoding "UTF-8"))
+      (open-output-string)))
+
+  (define (get-messages)
+    (let* ((str (get-output-string message-port))
+           (lst (string-split str #\newline)))
+      ;; 'string-tokenize' swallows empty lines, which is not great,
+      ;; and 'string-split' doesn't distinguish between an empty line
+      ;; and this empty string, which is not great either.  So we hack
+      ;; our way the best we can.
+      (cond ((string-null? str)
+             '())
+            ;; If STR ends in \n, drop the trailing empty string since
+            ;; that would lead the client to print an extra newline.
+            ((string-suffix? "\n" str)
+             (drop-right lst 1))
+            (else lst))))
+
+  (parameterize ((%current-client-socket message-port))
+    (proc get-messages)))
+
+(define-syntax-rule (with-command-message-port command get-messages
+                                               body ...)
+  "Evaluate @var{command} and bind @var{get-messages} in the lexical extent of
+@var{body} to a thunk to fetch messages emitted while evaluating
+@var{command}."
+  (call-with-command-message-port command
+                                  (lambda (get-messages)
+                                    body ...)))
+
 (define (process-command command port)
   "Interpret COMMAND, a command sent by the user, represented as a
 <shepherd-command> object.  Send the reply to PORT."
   (match command
-    (($ <shepherd-command> the-action service-symbol (args ...)
+    (($ <shepherd-command> version the-action service-symbol (args ...)
                            directory)             ;ignored
 
      ;; We have to catch `quit' so that we can send the terminator
      ;; line to herd before we actually quit.
      (catch 'quit
        (lambda ()
-         (define message-port
-           (with-fluids ((%default-port-encoding "UTF-8"))
-             (open-output-string)))
-
-         (define (get-messages)
-           (let* ((str (get-output-string message-port))
-                  (lst (string-split str #\newline)))
-             ;; 'string-tokenize' swallows empty lines, which is not great,
-             ;; and 'string-split' doesn't distinguish between an empty line
-             ;; and this empty string, which is not great either.  So we hack
-             ;; our way the best we can.
-             (cond ((string-null? str)
-                    '())
-                   ;; If STR ends in \n, drop the trailing empty string since
-                   ;; that would lead the client to print an extra newline.
-                   ((string-suffix? "\n" str)
-                    (drop-right lst 1))
-                   (else lst))))
-
-         (parameterize ((%current-client-socket message-port))
+         (with-command-message-port command get-messages
            (guard (c ((service-error? c)
                       (write-reply (command-reply command #f
                                                   (condition->sexp c)
                                                   (get-messages))
                                    port)))
 
+             (define service
+               (or (lookup-service service-symbol)
+                   (raise (condition
+                           (&missing-service-error (name service-symbol))))))
+
              (define result
                (case the-action
-                 ((start) (apply start service-symbol args))
-                 ((stop) (apply stop service-symbol args))
-                 ((enforce) (apply enforce service-symbol args))
+                 ((start)
+                  ;; Return #f or SERVICE: clients expect the service sexp.
+                  (if (eq? 'running (service-status service))
+                      (begin
+                        (local-output (l10n "Service ~a is already running.")
+		                      (service-canonical-name service))
+                        service)
+                      (and (apply start-service service args)
+                           service)))
+                 ((stop)
+                  (if (service-stopped? service)
+                      '()
+                      (map service-canonical-name
+                           (apply stop-service service args))))
 
-                 ;; Actions which have the semantics of `action' are
-                 ;; handled there.
-                 (else (apply action service-symbol the-action args))))
+                 ;; XXX: This used to return a list of action results, on the
+                 ;; grounds that there could be several services called NAME.
+                 ;; Clients like 'herd' expect a list so now we return a
+                 ;; singleton.
+                 (else (list (apply perform-service-action
+                                    service the-action args)))))
 
              (write-reply (command-reply command result #f (get-messages))
                           port))))
@@ -558,7 +563,7 @@ would write them on the 'herd' command line."
     (if (eof-object? line)
 
         ;; Exit on `C-d'.
-        (stop root-service)
+        (stop-service root-service)
 
         (begin
           (match (string-tokenize line)
@@ -570,3 +575,7 @@ would write them on the 'herd' command line."
             (_
              (local-output (l10n "invalid command line") line)))
           (loop (read-line port))))))
+
+;; Local Variables:
+;; eval: (put 'with-command-message-port 'scheme-indent-function 2)
+;; End:
