@@ -1,5 +1,5 @@
 # GNU Shepherd --- Test basic communication capabilities.
-# Copyright © 2013-2014, 2016-2019, 2022 Ludovic Courtès <ludo@gnu.org>
+# Copyright © 2013-2014, 2016-2019, 2022-2023 Ludovic Courtès <ludo@gnu.org>
 # Copyright © 2016 Mathieu Lirzin <mthl@gnu.org>
 # Copyright © 2014 Alex Sassmannshausen <alex.sassmannshausen@gmail.com>
 #
@@ -37,37 +37,41 @@ trap "cat $log || true; rm -f $socket $conf $stamp $log;
 cat > "$conf"<<EOF
 (use-modules (srfi srfi-26))
 (register-services
- (make <service>
-   #:provides '(test)
-   #:start (lambda _
-             (call-with-output-file "$stamp"
-               (cut display "foo" <>))
-             #t)
-   #:stop  (lambda _
-             (delete-file "$stamp"))
-   #:respawn? #f)
- (make <service>
-   #:provides '(test-2)
-   #:requires '(test)
-   #:start (lambda _
-             (call-with-output-file "$stamp-2"
-               (cut display "bar" <>))
-             #t)
-   #:stop  (lambda _
-             (delete-file "$stamp-2"))
-   #:actions (make-actions (hi "Say hi."
-                               (lambda _
-                                 (display "start\n\nend\n")
-                                 #t))
-			   (fail "Fail." (const #f)))
-   #:respawn? #f)
- (make <service>
-   #:provides '(broken)
-   #:requires '()
-   #:start (lambda _
-             (mkdir "/this/throws/a/system/error"))
-   #:stop  (const #f)
-   #:respawn? #f))
+ (list (service
+	 '(test)
+	 #:start (lambda _
+		   (call-with-output-file "$stamp"
+		     (cut display "foo" <>))
+		   #t)
+	 #:stop  (lambda _
+		   (delete-file "$stamp"))
+	 #:respawn? #f)
+       (service
+	 '(test-2)
+	 #:requirement '(test)
+	 #:start (lambda _
+		   (call-with-output-file "$stamp-2"
+		     (cut display "bar" <>))
+		   #t)
+	 #:stop  (lambda _
+		   (delete-file "$stamp-2"))
+	 #:actions (actions (hi "Say hi."
+				(lambda _
+				  (display "start\n\nend\n")
+				  #t))
+			    (fail "Fail." (const #f)))
+	 #:respawn? #f)
+       (service
+	 '(spawn-with-system)
+	 #:start (make-system-constructor "echo starting from $PWD")
+	 #:stop (make-system-destructor "echo stopping from $PWD"))
+       (service
+	 '(broken)
+	 #:requirement '()
+	 #:start (lambda _
+		   (mkdir "/this/throws/a/system/error"))
+	 #:stop  (const #f)
+	 #:respawn? #f)))
 EOF
 
 rm -f "$pid"
@@ -83,12 +87,18 @@ test -S "$socket"
 pristine_status=`$herd status root` # Prep for 'reload' test.
 echo $pristine_status | grep -E '(Start.*root|Stop.*test)'
 
+$herd graph | grep '"test-2" -> "test"'
+
 $herd start test
 test -f "$stamp"
-$herd status test | grep started
+$herd status test | grep running
 
 $herd stop test
-! test -f "$stamp"
+test -f "$stamp" && false
+
+$herd log
+$herd log | grep "service test is running"
+$herd log | grep "service test is stopped"
 
 $herd status test | grep stopped
 
@@ -124,7 +134,7 @@ if $herd fail test-2; then false; else true; fi
 if $herd enable test-2 with extra arguments
 then false; else true; fi
 
-$herd status test-2 | grep started
+$herd status test-2 | grep running
 
 # Make sure extra arguments lead to an error.
 if $herd status test-2 something else that is useless
@@ -141,8 +151,15 @@ done
 if $herd an-action-that-does-not-exist root
 then false; else true; fi
 
-if $herd start broken
-then false; else true; fi
+# Check the behavior for a service whose 'start' method throws.
+$herd start broken && false
+$herd status broken | grep "stopped"
+
+# Check 'make-system-constructor' and 'make-system-destructor'.
+$herd start spawn-with-system
+$herd status spawn-with-system | grep running
+$herd stop spawn-with-system
+$herd status spawn-with-system | grep "stopped"
 
 # Wrong number of arguments for an action.
 if $herd status root foo bar baz;
@@ -174,7 +191,7 @@ test "`$herd status`" == "$pristine_status"
 mkdir -p "$confdir"
 cat > "$confdir/some-conf.scm" <<EOF
 (register-services
- (make <service>
+ (make <service>                ;deprecated
    #:provides '(test-loaded)
    #:start (const 'abc)
    #:stop (const #f)))
@@ -199,7 +216,7 @@ $herd unload root test-loaded
 # make sure that the running value is clamped before being sent over the wire.
 cat > "$confdir/some-conf.scm" <<EOF
 (register-services
- (make <service>
+ (make <service>               ;deprecated
    #:provides '(test-loaded)
    #:start (const (if #f #f))  ;#<undefined>
    #:stop (const #f)))
@@ -211,7 +228,7 @@ $herd status test-loaded | grep -i "running.*#<unspecified>"
 $herd stop test-loaded
 
 # Deregister 'test-loaded' via 'eval'.
-$herd eval root "(action root-service 'unload \"test-loaded\")"
+$herd eval root "(perform-service-action root-service 'unload \"test-loaded\")"
 if $herd status test-loaded
 then false; else true; fi
 
@@ -253,8 +270,22 @@ if $herd status | grep "Stopped:"
 then false; else true; fi
 $herd status | grep -e "+ root"
 
+# Check that the "Service * already running" message gets printed only when
+# explicitly starting an already-running service.
+if grep "already running" "$log"
+then false; else true; fi
+
+$herd start root | grep "already running"
+grep "already running" "$log"
+
 $herd stop root
-! kill -0 $shepherd_pid
+
+# The shepherd process should be done now, or real soon.
+if kill -0 $shepherd_pid
+then
+    sleep 3
+    kill -0 $shepherd_pid && false
+fi
 
 test -f "$log"
 
@@ -279,15 +310,18 @@ while ! test -f "$pid" ; do sleep 0.3 ; done
 # Launch a service from $confdir/shepherd/init.scm.
 $herd start test
 test -f "$stamp"
-$herd status test | grep started
+$herd status test | grep running
 
 $herd stop test
-! test -f "$stamp"
+test -f "$stamp" && false
 
 shepherd_pid="`cat $pid`"
 
 $herd stop root
-! kill -0 $shepherd_pid
+if kill -0 $shepherd_pid
+then
+    sleep 3
+    kill -0 $shepherd_pid && false
+fi
 
-rm -rf $confdir
-rm -rf $datadir
+rm -rf "$confdir" "$datadir"
