@@ -1,5 +1,5 @@
 # GNU Shepherd --- Ensure file descriptors are not leaked to children.
-# Copyright © 2022 Ludovic Courtès <ludo@gnu.org>
+# Copyright © 2022, 2023 Ludovic Courtès <ludo@gnu.org>
 #
 # This file is part of the GNU Shepherd.
 #
@@ -32,6 +32,9 @@ herd="herd -s $socket"
 trap "cat $log || true; rm -f $socket $conf $log $fd_count $c_file $exe;
       test -f $pid && kill \`cat $pid\` || true; rm -f $pid" EXIT
 
+# GNU/Hurd lacks /proc/self/fd so far.
+[ -d /proc/self/fd ] || exit 77
+
 cat > "$c_file" <<EOF
 /* This program counts its own open file descriptors and writes
    that number to $fd_count.  It's more reliable than using the
@@ -43,6 +46,8 @@ cat > "$c_file" <<EOF
 #include <sys/types.h>
 #include <dirent.h>
 #include <string.h>
+#include <sys/socket.h>
+#include <assert.h>
 
 int
 main (int argc, char *argv[])
@@ -52,6 +57,14 @@ main (int argc, char *argv[])
   size_t count;
   FILE *log;
 
+  if (getenv ("LISTEN_FDS") != NULL)  /* systemd */
+    {
+      struct sockaddr_storage address;
+      socklen_t len = sizeof address;
+      int fd = accept (3, (struct sockaddr *) &address, &len);
+      assert (fd >= 0);
+    }
+
   dir = opendir ("/proc/self/fd");
   chdir ("/proc/self/fd");
   for (count = 0, ent = NULL; ent = readdir (dir), ent != NULL; )
@@ -82,33 +95,33 @@ EOF
 
 cat > "$conf" <<EOF
 (register-services
- (make <service>
-   #:provides '(system-ctor)
-   #:start (make-system-constructor "$exe")
-   #:stop  (const #f)
-   #:one-shot? #t)
- (make <service>
-   #:provides '(forkexec-ctor)
-   #:start (make-forkexec-constructor '("$(type -P sleep)" "100"))
-   #:stop (make-kill-destructor))
- (make <service>
-   #:provides '(inetd-ctor)
-   #:start (make-inetd-constructor '("$exe")
-                                   (list
-                                    (endpoint (make-socket-address
-                                               AF_INET
-                                               INADDR_LOOPBACK
-                                               5555))))
-   #:stop  (make-inetd-destructor))
- (make <service>
-   #:provides '(systemd-ctor)
-   #:start (make-systemd-constructor '("$exe")
-				     (list
-				      (endpoint (make-socket-address
-						 AF_INET
-						 INADDR_LOOPBACK
-						 5556))))
-   #:stop  (make-systemd-destructor)))
+ (list (service
+	 '(system-ctor)
+	 #:start (make-system-constructor "$exe")
+	 #:stop  (const #f)
+	 #:one-shot? #t)
+       (service
+	 '(forkexec-ctor)
+	 #:start (make-forkexec-constructor '("$(type -P sleep)" "100"))
+	 #:stop (make-kill-destructor))
+       (service
+	 '(inetd-ctor)
+	 #:start (make-inetd-constructor '("$exe")
+					 (list
+					  (endpoint (make-socket-address
+						     AF_INET
+						     INADDR_LOOPBACK
+						     5555))))
+	 #:stop  (make-inetd-destructor))
+       (service
+	 '(systemd-ctor)
+	 #:start (make-systemd-constructor '("$exe")
+					   (list
+					    (endpoint (make-socket-address
+						       AF_INET
+						       INADDR_LOOPBACK
+						       5556))))
+	 #:stop  (make-systemd-destructor))))
 EOF
 
 rm -f "$pid" "$fd_count"
@@ -170,13 +183,13 @@ do
     connect_to_server 5555
     test $(cat "$fd_count") -eq 4
 
-    # Spawn the systemd service by starting it (it actually immediately stops
-    # instead of calling 'accept', but it doesn't matter).  This one must have
-    # 4 open file descriptors, the 4th one being the socket.
+    # Spawn the systemd service by starting it.  This one must have
+    # 5 open file descriptors: one for the listening socket, and one for the
+    # accepted client connection (plus one for /proc/self/fd).
     $herd enable systemd-ctor
     $herd start systemd-ctor
     connect_to_server 5556
-    test $(cat "$fd_count") -eq 5
+    test $(cat "$fd_count") -eq 6
     $herd stop systemd-ctor
 
     $herd restart forkexec-ctor
