@@ -1,5 +1,5 @@
 # GNU Shepherd --- Test the logging capabilities of 'make-forkexec-constructor'.
-# Copyright © 2022 Ludovic Courtès <ludo@gnu.org>
+# Copyright © 2022, 2023 Ludovic Courtès <ludo@gnu.org>
 #
 # This file is part of the GNU Shepherd.
 #
@@ -48,25 +48,25 @@ cat > "$conf"<<EOF
   '("$SHELL" "$service_script"))
 
 (register-services
- (make <service>
-   ;; Service with built-in logging.
-   #:provides '(test-builtin-logging)
-   #:start (make-forkexec-constructor %command
-                                      #:pid-file "$PWD/$service_pid")
-   #:stop  (make-kill-destructor)
-   #:respawn? #f)
+ (list (service
+	 ;; Service with built-in logging.
+	 '(test-builtin-logging)
+	 #:start (make-forkexec-constructor %command
+					    #:pid-file "$PWD/$service_pid")
+	 #:stop  (make-kill-destructor)
+	 #:respawn? #f)
 
- (make <service>
-   ;; Service with built-in logging.
-   #:provides '(test-file-logging)
-   #:start (make-forkexec-constructor %command
-                                      #:log-file "$PWD/$service_log"
-                                      #:pid-file "$PWD/$service_pid")
-   #:stop  (make-kill-destructor)
-   #:respawn? #f))
+       (service
+	 ;; Service with built-in logging.
+	 '(test-file-logging)
+	 #:start (make-forkexec-constructor %command
+					    #:log-file "$PWD/$service_log"
+					    #:pid-file "$PWD/$service_pid")
+	 #:stop  (make-kill-destructor)
+	 #:respawn? #f)))
 
 ;; Start it upfront to make sure the logging fiber works.
-(start 'test-file-logging)
+(start-service (lookup-service 'test-file-logging))
 EOF
 
 rm -f "$pid"
@@ -78,7 +78,7 @@ while ! test -f "$pid" ; do sleep 0.3 ; done
 shepherd_pid="`cat $pid`"
 
 cat "$service_log"
-$herd status test-file-logging | grep started
+$herd status test-file-logging | grep running
 for message in "STARTING" "STARTED" "café" "latin1 garbage: .* alors"
 do
     grep -E '^2[0-9]{3}-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2} '"$message" "$service_log"
