@@ -1,5 +1,5 @@
 # GNU Shepherd --- Test respawnable services.
-# Copyright © 2013, 2014, 2016 Ludovic Courtès <ludo@gnu.org>
+# Copyright © 2013, 2014, 2016, 2023 Ludovic Courtès <ludo@gnu.org>
 #
 # This file is part of the GNU Shepherd.
 #
@@ -63,25 +63,26 @@ function assert_killed_service_is_respawned
 
 cat > "$conf"<<EOF
 (register-services
- (make <service>
-   #:provides '(test1)
-   #:start (make-forkexec-constructor
-	    '("$SHELL" "-c"
-	      "echo \$\$ > $PWD/$service1_pid ; while true ; do sleep 1 ; done"))
-   #:stop  (make-kill-destructor)
-   #:respawn? #t)
- (make <service>
-   #:provides '(test2)
-   #:start (make-forkexec-constructor
-            ;; The 'sleep' below is just to make it more likely
-            ;; that synchronization issues in handling #:pid-file
-            ;; would be caught.
-	    '("$SHELL" "-c"
-	      "sleep 0.7 ; echo \$\$ > $PWD/$service2_pid ; while true ; do sleep 1 ; done")
-            #:pid-file "$PWD/$service2_pid")
-   #:stop  (make-kill-destructor)
-   #:respawn? #t))
-(start 'test1)
+ (list
+  (service
+    '(test1)
+    #:start (make-forkexec-constructor
+	     '("$SHELL" "-c"
+	       "echo \$\$ > $PWD/$service1_pid ; while true ; do sleep 1 ; done"))
+    #:stop  (make-kill-destructor)
+    #:respawn? #t)
+  (service
+    '(test2)
+    #:start (make-forkexec-constructor
+	     ;; The 'sleep' below is just to make it more likely
+	     ;; that synchronization issues in handling #:pid-file
+	     ;; would be caught.
+	     '("$SHELL" "-c"
+	       "sleep 0.7 ; echo \$\$ > $PWD/$service2_pid ; while true ; do sleep 1 ; done")
+	     #:pid-file "$PWD/$service2_pid")
+    #:stop  (make-kill-destructor)
+    #:respawn? #t)))
+(start-service (lookup-service 'test1))
 EOF
 
 rm -f "$pid"
@@ -95,10 +96,10 @@ shepherd_pid="`cat $pid`"
 kill -0 $shepherd_pid
 test -S "$socket"
 $herd status
-$herd status test1 | grep started
+$herd status test1 | grep running
 
 $herd start test2
-$herd status test2 | grep started
+$herd status test2 | grep running
 
 # When 'herd start test2' returns, the PID file must already be created.
 test -f "$service2_pid"
@@ -124,8 +125,8 @@ pid="`cat "$service1_pid"`"
 rm "$service1_pid"
 $herd stop test1
 $herd status test1 | grep stopped
-! test -f "$service1_pid"
-! kill -0 "$pid"
+test -f "$service1_pid" && false
+kill -0 "$pid" && false
 
 cat $service2_pid
 $herd stop root
