@@ -1,5 +1,5 @@
 # GNU Shepherd --- Test whether 'system*' is blocking.
-# Copyright © 2022 Ludovic Courtès <ludo@gnu.org>
+# Copyright © 2022-2023 Ludovic Courtès <ludo@gnu.org>
 #
 # This file is part of the GNU Shepherd.
 #
@@ -34,27 +34,33 @@ script="while [ ! -f $PWD/$stamp ] ; do sleep 0.1 ; done ; exit \$(cat $PWD/$sta
 
 cat > "$conf" <<EOF
 (register-services
- (make <service>
-   #:provides '(test)
-   #:start (lambda _
-             (list 'exit-code
-                   (status:exit-val
-                    (system* "$SHELL" "-c" "$script"))))
-   #:stop  (lambda _
-             (system* "$SHELL" "-c" "echo STOPPING")
-             (delete-file "$stamp"))
-   #:respawn? #f)
- (make <service>
-   #:provides '(test-with-respawn)
-   #:start (make-forkexec-constructor
-             (list "$SHELL" "-cex"
-                   "[ ! -f $PWD/$stamp ] ; touch $PWD/$stamp ; sleep 60"))
-   #:stop  (lambda (pid)
-             (and (zero? (system* "$(type -P kill)" (number->string pid)))
-                  (begin
-                    (delete-file "$stamp")
-                    #f)))
-   #:respawn? #t))
+ (list (service
+	 '(test)
+	 #:start (lambda _
+		   (list 'exit-code
+			 (status:exit-val
+			  (system* "$SHELL" "-c" "$script"))))
+	 #:stop  (lambda _
+		   (system* "$SHELL" "-c" "echo STOPPING")
+		   (delete-file "$stamp"))
+	 #:respawn? #f)
+       (service
+	 '(test-command-not-found)
+	 #:start (lambda _
+		   (zero? (system* "this command does not exist")))
+	 #:stop  (const #f)
+	 #:respawn? #f)
+       (service
+	 '(test-with-respawn)
+	 #:start (make-forkexec-constructor
+		   (list "$SHELL" "-cex"
+			 "[ ! -f $PWD/$stamp ] ; touch $PWD/$stamp ; sleep 60"))
+	 #:stop  (lambda (pid)
+		   (and (zero? (system* "$(type -P kill)" (number->string pid)))
+			(begin
+			  (delete-file "$stamp")
+			  #f)))
+	 #:respawn? #t)))
 EOF
 
 rm -f "$pid"
@@ -70,8 +76,8 @@ kill -0 $shepherd_pid
 # 'herd start' will block until the script exits...
 $herd start test &
 
-# ... so at this point the service is stopped.
-$herd status test | grep "stopped"
+# ... so at this point the service is starting or about to start.
+$herd status test | grep -E "(starting|stopped)"
 
 # Touch $stamp.  The shell script passed to 'system*' should complete shortly
 # after that.
@@ -80,7 +86,7 @@ echo 123 > "$stamp"
 n=0
 while [ $n -lt 20 ]
 do
-    if $herd status test | grep "started"
+    if $herd status test | grep running
     then
 	break
     else
@@ -88,27 +94,40 @@ do
 	sleep 1
     fi
 done
-$herd status test | grep "started"
+$herd status test | grep running
 $herd status test | grep "exit-code 123"
 
 $herd stop test
-! test -f "$stamp"
+test -f "$stamp" && false
 grep "STOPPING" "$log"
 
+# This service uses 'system*' but the command is not found.
+$herd start test-command-not-found && false
+$herd status test-command-not-found
+$herd status test-command-not-found | grep "stopped"
+
 # What about a service with a custom 'stop' procedure that uses 'system*'?
 # Stopping the service should not trigger the respawn machinery.
 $herd start test-with-respawn
-$herd status test-with-respawn | grep "started"
+$herd status test-with-respawn | grep running
 $herd stop test-with-respawn
 $herd status test-with-respawn | grep "stopped"
 
 for i in `seq 1 5`
 do
     $herd restart test-with-respawn
-    $herd status test-with-respawn | grep "started"
+    $herd status test-with-respawn | grep running
 done
+$herd stop test-with-respawn
+
+# What happens when we cause the process monitor to throw an exception while
+# trying to fork?  The process monitor fiber should remain alive.
+$herd eval root "(setrlimit 'nproc 1 1)"
+$herd start test && false
+$herd status test
+$herd status test | grep "stopped"
 
 $herd stop root
 
 # Make sure 'shutdown-services' did its job.
-! test -f "$stamp"
+if test -f "$stamp"; then false; else true; fi
